/*
 * Part of the JavaChan project.
 * (Java ikaChan)
 *
 *        DO WHATEVER YOU WANT TO PUBLIC LICENSE 
 *                    Version 2, December 2004 
 *
 * Modified by gamemanj <asdd2808@gmail.com> to remove expletives
 * Original copyright (C) 2004 Sam Hocevar <sam@hocevar.net> 
 *
 * Everyone is permitted to copy and distribute verbatim or modified 
 * copies of this license document, and changing it is allowed as long 
 * as the name is changed. 
 *
 *            DO WHATEVER YOU WANT TO PUBLIC LICENSE 
 *   TERMS AND CONDITIONS FOR COPYING, DISTRIBUTION AND MODIFICATION 
 *
 *  0. You just DO WHATEVER YOU WANT TO.
 */
package javachan.games;

import java.io.FileOutputStream;
import javachan.Game;
import javachan.GameState;
import javachan.GameStateNPChar.NPCCharacter;
import javachan.IGame;
import gabien.IGrInDriver;
import gabien.IGrInDriver.IImage;
import javachan.Rect;
import javachan.blocks.BasicBlockType;
import javachan.blocks.DirectionBlockType;
import javachan.games.editorui.UIButton;
import javachan.games.editorui.UIHorizontalList;
import javachan.games.editorui.UIHorizontalList.ListEntry;
import javachan.games.editorui.UINumberBox;
import javachan.games.editorui.UIPanel;
import javachan.npcs.BasicNPCType;

/**
 * 
 * Yep,there's a LOT of code here.
 * Note that the layout code is still pretty horrible.
 * However,it's a lot less horrible than it was before.
 * ('before' means a time where everything was done by insanely long expressions.
 *   Occasionally I'd use a javachan.Rect to keep stuff in shape.
 *   See src folder of v1 to v6.1,or if you don't have it,ask for it via PM.
 *   v6 is missing due to the fact that v6.1 was correcting a packaging error.)
 * 
 * @author matt
 */
//NEW(1x GUI,as recommended by andwhyisit.)_____________________________________
//Note the NPC editor layout is NOW like this:
/*NPCW           NPCTW               MPW
 * 128 | width-(MPW+NPCW) | 96
 * ----------------------------------
 * so NPCTW now makes the sacrifice.
 * and vertically:
 * MPH 64
 */
//OLD_____________________________________
//Note the NPC editor layout is like this:
/*NPCW NPCTW    MPW
 * 64 | 216-NPCW | width-(NPCW+NPCTW)
 * ----------------------------------
 * 
 * and vertically:
 * MPH 32
 */

public class EditorGame implements IGame {

    private static class TileListEntry implements ListEntry {
        public int Type,Subtype;
        public boolean NoSubtypes;
        public BasicBlockType BlockType;
        public double InternalClock=0;//kludge:we don't receive TSS time,we only get delta
        public TileListEntry(int A,int B,boolean C,BasicBlockType[] BT) {
            BlockType=BT[A];
            Type=A;
            Subtype=B;
            NoSubtypes=C;
        }

        @Override
        public boolean IsEqualTo(ListEntry le) {
            if (((TileListEntry)le).Type==Type)
            if (NoSubtypes||((TileListEntry)le).NoSubtypes||(((TileListEntry)le).Subtype==Subtype))
                return true;
            return false;
        }

        @Override
        public void Draw(int centerx, int centery, double DeltaTime, IGrInDriver igd) {
            
            InternalClock+=DeltaTime;
            if (BlockType!=null)
            {
            igd.BlitImage(160+(Type*8), 48, 16, 16,centerx-8, centery-24, igd.GetImage("TileEditor.png", 0, 0, 0));
            if (BlockType instanceof DirectionBlockType)
            {
            igd.BlitImage(16+(16*Subtype), 32, 16, 16,centerx-8, centery-8, igd.GetImage("TileEditor.png", 0, 0, 0));
            }
            BlockType.Draw(0,centerx-8,centery-8,Subtype, igd,InternalClock);
            BlockType.Draw(1,centerx-8,centery-8,Subtype, igd,InternalClock);
            } else {
            igd.BlitImage(0, 32, 16, 16,centerx-8,centery-8, igd.GetImage("TileEditor.png", 0, 0, 0));
            }
        }
    }
    
    private static class NPCListEntry implements ListEntry {
        public int Type,Subtype;
        public boolean NoSubtypes;
        public GameState gs;
        public NPCListEntry(int A,int B,boolean C,GameState gs) {
            Type=A;
            Subtype=B;
            NoSubtypes=C;
            this.gs=gs;
        }

        @Override
        public boolean IsEqualTo(ListEntry le) {
            if (((NPCListEntry)le).Type==Type)
            if (NoSubtypes||((NPCListEntry)le).NoSubtypes||(((NPCListEntry)le).Subtype==Subtype))
                return true;
            return false;
        }

        @Override
        public void Draw(int centerx, int centery, double DeltaTime, IGrInDriver igd) {
            NPCCharacter npcc=new NPCCharacter();
            npcc.Exists=true;
            npcc.Type=Type;
            npcc.OX=Subtype;
            BasicNPCType bnt=gs.npcs.npctypes[npcc.Type];
            bnt.Draw(centerx-8,centery-8, DeltaTime,npcc,gs,igd);
        }
    }
    //This is for the really nice NPC scroll thing.
    //For IdioticBaka1824
    public final int[] NPCTypes =          {0, 1, 1, 1, 1, 1, 2, 3, 4, 5,6,6,6, 7, 8, 9};
    public final int[] NPCSubTypes =       {0, 0, 1, 2, 3, 4, 0, 0, 0, 0,0,1,2, 0, 0, 0};
    public final boolean[] NPCSubTypeLess = {true,false,true,true,true,true,false,true,true,true};
    //Part of the BCM update(BCM:Big Code Merge-unifying horizontal list code,finally,since 2014).
    public final int[] TILESTypes =          {0,2,2,2,2,4,6,6,6,6,8,8,8,8,8,8,8,8,10,10,14};
    public final int[] TILESSubTypes =       {0,0,1,2,3,0,0,1,2,3,0,1,2,3,4,5,6,7,0,1,0};
    public final boolean[] TILESSubTypeLess= {true,false,false,false,false,false,false,false,false,false,false,false,false,false,false,false};
    
    public GameState gs;
    public int Tab = 0;
    //LEVEL 1 ELEMENTS
    public UIPanel NPCTab=new UIPanel(),TileTab=new UIPanel(),Common=new UIPanel();
    //LEVEL 2 ELEMENTS
    //common
    public UIButton SaveButton=new UIButton(),TabSwitch=new UIButton();
    //npc
    public UIPanel NumbersPanel=new UIPanel();
    public UIHorizontalList NPCS=new UIHorizontalList();
    //tiles
    public UINumberBox Tile_Number=new UINumberBox();
    public UIHorizontalList TILES=new UIHorizontalList();
    //LEVEL 3 ELEMENTS
    //npc
    public UIPanel NPC_CheckBox=new UIPanel();
    public UIButton NPC_CollisionType=new UIButton();
    public UINumberBox NPC_EventID=new UINumberBox(),NPC_NPCID=new UINumberBox(),NPC_X=new UINumberBox(),NPC_Y=new UINumberBox();
    //PX,PY:Camera position(as draw offset)
    public double PX = 0, PY = 0;
    //If 0 or under,then draw save normally,else draw it offset a bit so the user knows it saved
    public double SavePressTime = 0;

    public EditorGame(GameState ggs) {
        
        gs = ggs;
        NPCS.Types=new UIHorizontalList.ListEntry[NPCTypes.length];
        for (int x=0;x<NPCTypes.length;x++)
        {
            NPCS.Types[x]=new NPCListEntry(NPCTypes[x],NPCSubTypes[x],NPCSubTypeLess[NPCTypes[x]],gs);
        }
        TILES.Types=new UIHorizontalList.ListEntry[TILESTypes.length];
        for (int x=0;x<TILESTypes.length;x++)
        {
            TILES.Types[x]=new TileListEntry(TILESTypes[x],TILESSubTypes[x],TILESSubTypeLess[TILESTypes[x]],gs.map.blocktypes);
        }
        NumbersPanel.SetBounds(new Rect(0,0,128,64));
        TILES.OnChange=new Runnable(){

            @Override
            public void run() {
                Tile_Number.Number=(((TileListEntry)TILES.Types[TILES.SelectedItem]).Type*16)+(((TileListEntry)TILES.Types[TILES.SelectedItem]).Subtype);
            }
        };
        NPC_CheckBox.SetBounds(new Rect(2,10,16,8));
        NPC_CheckBox.ImageX=304;
        NPC_CollisionType.SetBounds(new Rect(40,34,48,14));
        NPC_CollisionType.OnClick=new Runnable() {

            @Override
            public void run() {
                gs.npcs.chars[NPC_NPCID.Number].CollisionType++;
                if (gs.npcs.chars[NPC_NPCID.Number].CollisionType>=4)
                    gs.npcs.chars[NPC_NPCID.Number].CollisionType=0;
            }
        };
        NPC_NPCID.SetBounds(new Rect(40,0,64,16));
        NPCS.OnChange=new Runnable() {

            @Override
            public void run() {
                gs.npcs.chars[NPC_NPCID.Number].Type=((NPCListEntry)NPCS.Types[NPCS.SelectedItem]).Type;
                gs.npcs.chars[NPC_NPCID.Number].OX=((NPCListEntry)NPCS.Types[NPCS.SelectedItem]).Subtype;
            }
        };
        NPC_NPCID.OnEdit=new Runnable() {

            @Override
            public void run() {
                if (NPC_NPCID.Number>=100) NPC_NPCID.Number=0;
                if (NPC_NPCID.Number<0) NPC_NPCID.Number=0;
                NPC_EventID.Number=gs.npcs.chars[NPC_NPCID.Number].EventID;
                NPC_X.Number=(int)gs.npcs.chars[NPC_NPCID.Number].X;
                NPC_Y.Number=(int)gs.npcs.chars[NPC_NPCID.Number].Y;
                NPCListEntry cmp=new NPCListEntry(gs.npcs.chars[NPC_NPCID.Number].Type,(int)gs.npcs.chars[NPC_NPCID.Number].OX,false,gs);
                for (int p=0;p<NPCS.Types.length;p++)
                {
                    if (cmp.IsEqualTo(NPCS.Types[p])) NPCS.SelectedItem=p;
                }
            }
        };
        NPC_EventID.SetBounds(new Rect(40,16,64,16));
        NPC_EventID.OnEdit=new Runnable() {

            @Override
            public void run() {
                if (NPC_EventID.Number>=256) NPC_EventID.Number=0;
                if (NPC_EventID.Number<0) NPC_EventID.Number=0;
                gs.npcs.chars[NPC_NPCID.Number].EventID=NPC_EventID.Number;
            }
        };
        NPC_X.SetBounds(new Rect(28,50,64,16));
        NPC_X.OnEdit=new Runnable() {

            @Override
            public void run() {
                gs.npcs.chars[NPC_NPCID.Number].X=NPC_X.Number;
            }
        };
        NPC_Y.SetBounds(new Rect(94,50,64,16));
        NPC_Y.OnEdit=new Runnable() {

            @Override
            public void run() {
                gs.npcs.chars[NPC_NPCID.Number].Y=NPC_Y.Number;
            }
        };
        NPC_NPCID.OnEdit.run();

        NPCTab.Elements.add(Common);
        NPCTab.Elements.add(NumbersPanel);
        NPCTab.Elements.add(NPCS);
        NumbersPanel.Elements.add(NPC_EventID);
        NumbersPanel.Elements.add(NPC_NPCID);
        NumbersPanel.Elements.add(NPC_X);
        NumbersPanel.Elements.add(NPC_Y);
        NumbersPanel.Elements.add(NPC_CheckBox);
        NumbersPanel.Elements.add(NPC_CollisionType);

        TileTab.Elements.add(Common);
        TILES.Elements.add(Tile_Number);
        TileTab.Elements.add(TILES);
        SaveButton.SetBounds(new Rect(0,0,32,32));
        SaveButton.OnClick=new Runnable()
                {

            @Override
            public void run() {
                try {
                    FileOutputStream fos=new FileOutputStream("Pbm/Map1.pbm");
                    gs.map.SaveBitmap(fos);
                    fos.close();
                    fos=new FileOutputStream("NPChar.dat");
                    gs.npcs.Save(fos);
                    fos.close();
                } catch (Exception e)
                {
                }
            }
                };
                TabSwitch.OnClick=new Runnable()
                {

            @Override
            public void run() {
                Tab^=1;
            }
                };

        TabSwitch.SetBounds(new Rect(32,0,64,32));
        Common.Elements.add(SaveButton);
        Common.Elements.add(TabSwitch);
    }
    @Override
    public void Run(double DeltaTime, double Time, Game g) {
        //---------------------------------START MY LAYOUT CODE
        Common.SetBounds(new Rect(g.graphics.GetWidth()-96,0,96,64));
        NPCTab.SetBounds(new Rect(0,0,g.graphics.GetWidth(),64));
        TileTab.SetBounds(new Rect(0,0,g.graphics.GetWidth(),64));
        Tile_Number.SetBounds(new Rect(TileTab.GetBounds().width-64,TileTab.GetBounds().height-16,64,16));
        //128(NumbersPanel Size)+96(Common Size)
        NPCS.SetBounds(new Rect(128,0,g.graphics.GetWidth()-(128+96),64));
        //96(Common Size)
        TILES.SetBounds(new Rect(0,0,g.graphics.GetWidth()-96,64));
        IImage edit = g.graphics.GetImage("TileEditor.png", 0, 0, 0);
        NPC_CollisionType.Image=NPC_CheckBox.Image=NumbersPanel.Image=TabSwitch.Image=SaveButton.Image=edit;
        NPC_CollisionType.ImageX=448;
        NPC_CollisionType.ImageY=(gs.npcs.chars[NPC_NPCID.Number].CollisionType*16)+2;
        NumbersPanel.ImageX=320;
        
        if (gs.npcs.chars[NPC_NPCID.Number].Exists)
        {
        NPC_CheckBox.ImageY=0;
        } else {
        NPC_CheckBox.ImageY=8;
        }
        TabSwitch.ImageX=32+(64*Tab);
        //--------------------------------END MY LAYOUT CODE
        g.graphics.Clear(0, 0, 0);
        int cmx = g.graphics.GetMouseX();
        int cmy = g.graphics.GetMouseY();
        int cmb = g.graphics.GetMouseButton();
        int Mx = (int) Math.floor((cmx - PX) / 16);//in !tiles!
        int My = (int) Math.floor((cmy - PY) / 16);//in !tiles!
            if (g.graphics.GetMouseJustDown())
            {
                if (cmy<64)
                {
                g.graphics.SetTypeBuffer("");
                if (Tab==0)
                {
                NPCTab.HandleClick(cmx, cmy);
                } else {
                TileTab.HandleClick(cmx, cmy);
                }
                } else {
                if (Tab==0)
                {
                    if (cmb==1)
                    {
                gs.npcs.chars[NPC_NPCID.Number].X=Mx;
                gs.npcs.chars[NPC_NPCID.Number].Y=My;
                gs.npcs.chars[NPC_NPCID.Number].Exists=true;
                    } 
                    if (cmb==3)
                    {
                gs.npcs.chars[NPC_NPCID.Number].Exists=false;
                    }
                } else {
                    if (cmb==1)
                    {
                    if (Mx>=0)
                        if (My>=0)
                            if (Mx<gs.map.Width)
                                if (My<gs.map.Height)
                                    gs.map.data[Mx+(My*gs.map.Width)]=Tile_Number.Number;
                    } else {
                    if (Mx>=0)
                        if (My>=0)
                            if (Mx<gs.map.Width)
                                if (My<gs.map.Height)
                                    gs.map.data[Mx+(My*gs.map.Width)]=0;
                    }
                }
                }
            }
            if (g.graphics.GetMouseDown())
            {
                if (cmy<64)
                {
                    if (cmb==1)
                    {
                if (Tab==0)
                {
                NPCTab.HandleDrag(cmx, cmy);
                } else {
                TileTab.HandleDrag(cmx, cmy);
                }
                    }
                } else {
                if (Tab==0)
                {
                } else {
                }
                }
            }
        if (g.graphics.IsKeyJustPressed(Keymap.VK_SWIM))
        {
            if (Tab==0)
            {
                //low on code duplication version of loop I mentioned.
                //while loops are way more flexible.
                int op=NPC_NPCID.Number;//simple way of making shorthand,
                                        //and there's a 100% guarantee we won't infiniteloop
                                        //even if another thread monkeys around.
                int p=(op+1)%gs.npcs.chars.length;
                boolean noblank=false;
                while (p!=op)
                {
                    NPCCharacter npcc=gs.npcs.chars[p];
                    if (npcc.Exists)
                    if (new Rect(Mx*16,My*16,16,16).intersects(new Rect((int)(npcc.X*16),(int)(npcc.Y*16),16,16)))
                    {
                        noblank=true;
                        NPC_NPCID.Number=p;
                        NumbersPanel.Selected=null;
                        NPC_NPCID.OnEdit.run();
                        break;
                    }
                    p++;
                    if (p>=gs.npcs.chars.length)
                        p=0;
                }
                if (!noblank)
                {
                    for (int pp=0;pp<gs.npcs.chars.length;pp++)
                    {
                       if (!gs.npcs.chars[pp].Exists)
                       {
                        NPC_NPCID.Number=pp;
                        NumbersPanel.Selected=null;
                        NPC_NPCID.OnEdit.run();
                        break;
                       }
                    }
                }
            }
        }
        if (g.graphics.IsKeyDown(IGrInDriver.VK_DOWN)) {
            PY -= DeltaTime * 256;
        }
        if (g.graphics.IsKeyDown(IGrInDriver.VK_UP)) {
            PY += DeltaTime * 256;
        }
        if (g.graphics.IsKeyDown(IGrInDriver.VK_LEFT)) {
            PX += DeltaTime * 256;
        }
        if (g.graphics.IsKeyDown(IGrInDriver.VK_RIGHT)) {
            PX -= DeltaTime * 256;
        }
        gs.RenderBack((int) PX, (int) PY, Time, g.graphics);
        gs.map.RenderTiles((int) PX, (int) PY, 0, Time, g.graphics);
        for (int xx=-1;xx<gs.map.Width+1;xx++)
        {
        for (int yy=-1;yy<gs.map.Height+1;yy++)
        {
        if (new Rect(0,0,gs.map.Width,gs.map.Height).contains(xx, yy))
            continue;
        g.graphics.BlitImage(288, 0, 16, 16, ((xx * 16) + ((int) PX)), ((yy * 16) + ((int) PY)), edit);
        }
        }
        gs.npcs.Render((int) PX, (int) PY, Time,gs,g.graphics);
        gs.RenderPC((int) PX, (int) PY, 0, g.graphics,Time);
        gs.map.RenderTiles((int) PX, (int) PY, 1, Time, g.graphics);
        g.graphics.BlitBCKImage(0, 32, 16, 16, ((Mx * 16) + ((int) PX)), ((My * 16) + ((int) PY)), edit);
        if (Tab==0)
        {
        g.graphics.BlitBCKImage(32, 16, 16, 16, (int)(gs.npcs.chars[NPC_NPCID.Number].X*16)+(int)PX,(int)(gs.npcs.chars[NPC_NPCID.Number].Y*16)+(int)PY, g.graphics.GetImage("Pbm/Cursor.pbm", 0,0,0));
        for (int pp=0;pp<gs.npcs.chars.length;pp++)
        {
        	if (gs.npcs.chars[pp].Exists)
            g.graphics.BlitBCKImage(32, 0, 16, 16, (int)(gs.npcs.chars[pp].X*16)+(int)PX,(int)(gs.npcs.chars[pp].Y*16)+(int)PY, g.graphics.GetImage("Pbm/Cursor.pbm", 0,0,0));
        }
        }
        DisplayNumber(Mx, 0, g.graphics.GetHeight() - 16, g.graphics,edit);
        DisplayNumber(My, 64, g.graphics.GetHeight() - 16, g.graphics,edit);
        if (Tab==0)
        {
            TileTab.Selected=null;
            NPCTab.UpdateAndDraw(0,0, DeltaTime, true, g.graphics);
        } else {
            NPCTab.Selected=null;
            TileTab.UpdateAndDraw(0,0, DeltaTime, true, g.graphics);
        }
        g.graphics.Flush();
    }

    public static void DisplayNumber(int NPC_NO, int i, int i0, IGrInDriver graphics,IImage edit) {
        String ss = "" + NPC_NO;
        int pp = 0;
        for (char c : ss.toCharArray()) {
            graphics.BlitImage((c - '0') * 8, 48, 8, 16, i + pp, i0,edit);
            pp += 9;
        }
    }
}
